/* 
 * HPCU Bi-Weekly Challenge: Hybrid Parallel Pan Water Cycle
 * Author: Aaron Weeden, Shodor, 2014
 * Last Modified: October 11, 2014, by Aaron Weeden
 */

/* import libraries */
#include <omp.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

/* create macro for exiting if MPI errors occur */
#define tryMPI(ret, str, rank) \
    if (ret != MPI_SUCCESS) { \
        fprintf(stderr, "Rank %d returned error for %s, exiting!\n", \
                rank, str); \
        exit(EXIT_FAILURE); \
    }

/* declare constants */
const double TEMPERATURE = 100.0;
const double EVAPORATION_PER_DEGREE = 0.05;
const double DEFAULT_CONDENSATION_FACTOR = 0.0;
const double DEFAULT_PRECIPITATION_FACTOR = 0.0;
const double INITIAL_PAN = 100.0;
const double INITIAL_COVER = 0.0;
const double INITIAL_AIR = 0.0;
const int NUM_TIME_STEPS = 10;
const int MPI_TAG = 0;

/* start program */
int main(int argc, char **argv) {

    /* declare variables */
    int mpi_rank;
    int mpi_size;
    int num_threads;
    double pan;
    double cover;
    double air;
    double condensation;
    double precipitation;
    double evaporation;
    double condensation_factor;
    double precipitation_factor;
    double *final_pan;
    double *final_cover;
    double *final_air;
    int t;
    int i;
    int j;

    /* set up mpi */
    tryMPI(
            MPI_Init(&argc, &argv),
            "MPI_Init",
            0
          );
    tryMPI(
            MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank),
            "MPI_Comm_rank",
            0
          );
    tryMPI(
            MPI_Comm_size(MPI_COMM_WORLD, &mpi_size),
            "MPI_Comm_size",
            mpi_rank
          );

    /* get number of threads */
#pragma omp parallel
    {
#pragma omp single
        {
            num_threads = omp_get_num_threads();
        }
    }

    /* allocate memory for arrays (enough space for each thread) */
    if ((final_pan = malloc(num_threads * sizeof(double))) == NULL) {
        fprintf(stderr,
                "Rank %d returned error for final_pan malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }
    if ((final_cover = malloc(num_threads * sizeof(double))) == NULL) {
        fprintf(stderr,
                "Rank %d returned error for final_cover malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }
    if ((final_air = malloc(num_threads * sizeof(double))) == NULL) {
        fprintf(stderr,
                "Rank %d returned error for final_air malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }

    /* initialize condensation factor */
    condensation_factor = mpi_size > 1
        ? (double)mpi_rank/(mpi_size-1)
        : DEFAULT_CONDENSATION_FACTOR;

    /* spawn OpenMP threads */
#pragma omp parallel private(precipitation_factor, pan, cover, air, t, \
        condensation, precipitation, evaporation)
    {
        /* initialize precipitation factor */
        precipitation_factor = num_threads > 1
            ? (double)omp_get_thread_num()/(num_threads-1)
            : DEFAULT_PRECIPITATION_FACTOR;

        /* initialize water amounts */
        pan = INITIAL_PAN;
        cover = INITIAL_COVER;
        air = INITIAL_AIR;

        /* run the simulation */
        for (t = 0; t < NUM_TIME_STEPS; t++) {
            condensation = condensation_factor * air;
            precipitation = precipitation_factor * cover;
            evaporation = TEMPERATURE * EVAPORATION_PER_DEGREE;
            pan += precipitation - evaporation;
            cover += condensation - precipitation;
            air += evaporation - condensation;

            /* make sure no amount is negative */
            if (pan < 0.0) pan = 0.0;
            if (cover < 0.0) cover = 0.0;
            if (air < 0.0) air = 0.0;
        }

        /* store the final water amounts in arrays */
        final_pan[omp_get_thread_num()] = pan;
        final_cover[omp_get_thread_num()] = cover;
        final_air[omp_get_thread_num()] = air;

        /* done spawning threads */
    }

    /* send arrays to rank 0 */
    tryMPI(
            MPI_Send(final_pan, num_threads, MPI_DOUBLE, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );
    tryMPI(
            MPI_Send(final_cover, num_threads, MPI_DOUBLE, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );
    tryMPI(
            MPI_Send(final_air, num_threads, MPI_DOUBLE, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );

    /* rank 0 only */
    if (mpi_rank == 0) {

        /* receive arrays from each process */
        for (i = 0; i < mpi_size; i++) {
            tryMPI(
                    MPI_Recv(final_pan, num_threads, MPI_DOUBLE, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );
            tryMPI(
                    MPI_Recv(final_cover, num_threads, MPI_DOUBLE, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );
            tryMPI(
                    MPI_Recv(final_air, num_threads, MPI_DOUBLE, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );

            /* print results */
            for (j = 0; j < num_threads; j++) {
                printf("%f %f %f %f %f\n",
                        mpi_size > 1
                        ? (double)i/(mpi_size-1)
                        : DEFAULT_CONDENSATION_FACTOR,
                        num_threads > 1
                        ? (double)j/(num_threads-1)
                        : DEFAULT_PRECIPITATION_FACTOR,
                        final_pan[j],
                        final_cover[j],
                        final_air[j]);
            }
        }
    }

    /* free memory for arrays */
    free(final_air);
    free(final_cover);
    free(final_pan);

    /* tear down mpi */
    tryMPI(
            MPI_Finalize(),
            "MPI_Finalize",
            mpi_rank
          );

    /* quit program */
    return 0;
}
